from typing import Type

from django.db.models import Q
from django.db.models.expressions import F
from django.db.models.query import QuerySet
from wbcore import serializers
from wbcore.serializers.serializers import Serializer
from wbhuman_resources.models.kpi import KPI, KPIHandler
from wbhuman_resources.serializers import KPIModelSerializer

from wbintegrator_office365.models import CallEvent


class NumberOfCallKPIKPISerializer(KPIModelSerializer):
    call_area = serializers.ChoiceField(
        default="all",
        choices=[("only_internal", "Only Internal"), ("only_external", "Only External"), ("all", "All")],
    )
    person_participates = serializers.BooleanField(
        default=True,
        label="Participants of call/Meeting ",
        help_text="Calls/Meeting considered are related to the participants",
    )
    person_created = serializers.BooleanField(
        default=True,
        label="Organizer of call/Meeting ",
        help_text="Calls/Meeting considered are related to the organizer",
    )

    def update(self, instance, validated_data):
        call_area = validated_data.get(
            "call_area",
            instance.additional_data["serializer_data"].get("call_area", "all"),
        )

        person_participates = validated_data.get(
            "person_participates",
            instance.additional_data["serializer_data"].get("person_participates", True),
        )
        person_created = validated_data.get(
            "person_created",
            instance.additional_data["serializer_data"].get("person_created", True),
        )

        additional_data = instance.additional_data
        additional_data["serializer_data"]["call_area"] = call_area
        additional_data["serializer_data"]["person_participates"] = person_participates
        additional_data["serializer_data"]["person_created"] = person_created

        additional_data["list_data"] = instance.get_handler().get_list_data(additional_data["serializer_data"])
        validated_data["additional_data"] = additional_data

        return super().update(instance, validated_data)

    class Meta(KPIModelSerializer.Meta):
        fields = (
            *KPIModelSerializer.Meta.fields,
            "call_area",
            "person_participates",
            "person_created",
        )


class NumberOfCallKPI(KPIHandler):
    def get_name(self) -> str:
        return "Number of Calls/Meeting"

    def get_serializer(self) -> Type[Serializer]:
        return NumberOfCallKPIKPISerializer

    def annotate_parameters(self, queryset: QuerySet[KPI]) -> QuerySet[KPI]:
        return queryset.annotate(
            call_area=F("additional_data__serializer_data__call_area"),
            person_participates=F("additional_data__serializer_data__person_participates"),
            person_created=F("additional_data__serializer_data__person_created"),
            person_assigned=F("additional_data__serializer_data__person_assigned"),
        )

    def get_list_data(self, serializer_data: dict) -> list[str]:
        return [
            f"Call Area: {serializer_data['call_area']}",
            f"Person Participates: {serializer_data['person_participates']}",
            f"Person Created: {serializer_data['person_created']}",
        ]

    def get_display_grid(self) -> list[list[str]]:
        return [["call_area"] * 2, ["person_created", "person_participates"]]

    def evaluate(self, kpi: "KPI", evaluated_person=None, evaluation_date=None) -> int:
        persons = (
            [evaluated_person.id] if evaluated_person else kpi.evaluated_persons.all().values_list("id", flat=True)
        )
        serializer_data = kpi.additional_data.get("serializer_data")
        qs = CallEvent.objects.filter(
            start__gte=kpi.period.lower,
            end__lte=evaluation_date if evaluation_date else kpi.period.upper,
        )
        if (call_area := serializer_data.get("call_area")) and (call_area != "all"):
            if call_area == "only_internal":
                qs = qs.filter(is_internal_call=True)
            elif call_area == "only_external":
                qs = qs.filter(is_internal_call=False)

        condition = None
        if serializer_data.get("person_created") or serializer_data.get("person_created") is None:
            condition = Q(organizer__tenant_user__profile__in=persons)
        if serializer_data.get("person_participates") or serializer_data.get("person_participates") is None:
            condition = (
                (condition | Q(participants__tenant_user__profile__in=persons))
                if condition
                else Q(participants__tenant_user__profile__in=persons)
            )
        if condition:
            qs = qs.filter(condition)
        return qs.distinct().count()



from rest_framework.reverse import reverse
from wbcore.metadata.configs.endpoints import EndpointViewConfig


class MixinEndpointConfig(EndpointViewConfig):
    def get_endpoint(self):
        return super().get_endpoint()

    def get_list_endpoint(self):
        return reverse(f"{self.view.get_model().get_endpoint_basename()}-list", request=self.request)

    def get_instance_endpoint(self):
        if self.instance:
            return None
        return super().get_instance_endpoint()

    def get_create_endpoint(self):
        return None

    def get_delete_endpoint(self):
        return None


class TenantUserEndpointConfig(MixinEndpointConfig):
    pass


class EventEndpointConfig(MixinEndpointConfig):
    pass


class CallUserEndpointConfig(MixinEndpointConfig):
    pass


class CallEventEndpointConfig(MixinEndpointConfig):
    pass


class SubscriptionEndpointConfig(MixinEndpointConfig):
    pass


class EventLogEndpointConfig(MixinEndpointConfig):
    pass


class EventLogEventEndpointConfig(MixinEndpointConfig):
    def get_endpoint(self, **kwargs):
        if event_id := self.view.kwargs.get("last_event_id", None):
            return reverse("wbintegrator_office365:event-eventlog-list", args=[event_id], request=self.request)
        return None



from wbcore.metadata.configs.titles import TitleViewConfig


class TenantUserTitleConfig(TitleViewConfig):
    def get_instance_title(self):
        return "Tenant User: {{tenant_id}}"

    def get_list_title(self):
        return "Tenant Users"

    def get_create_title(self):
        return "New Tenant User"


class CallUserTitleConfig(TitleViewConfig):
    def get_instance_title(self):
        return "Call User: {{_tenant_user.profile_str}}"

    def get_list_title(self):
        return "Call Users"

    def get_create_title(self):
        return "New Call User"


class CallEventTitleConfig(TitleViewConfig):
    def get_instance_title(self):
        return "Call Event: {{_event.id}}"

    def get_list_title(self):
        return "Call Events"

    def get_create_title(self):
        return "New Call Event"


class CallEventReceptionTimeTitleConfig(TitleViewConfig):
    def get_list_title(self):
        return "Call Events Reception Delay"


class CallEventSummaryGraphTitleConfig(TitleViewConfig):
    def get_list_title(self):
        return "Call Events Summary Graph"


class SubscriptionTitleConfig(TitleViewConfig):
    def get_instance_title(self):
        return "Subscription: {{id_str}}"



from celery import chain, shared_task
from django.db import models, transaction
from django.db.models import Q
from django.db.models.signals import post_delete, post_save
from django.dispatch import receiver
from wbcore.models import WBModel

from wbintegrator_office365.importer import MicrosoftGraphAPI


class Subscription(WBModel):
    class Meta:
        verbose_name = "Subscription"
        verbose_name_plural = "Subscriptions"

    class TypeResource(models.TextChoices):
        CALLRECORD = "CALLRECORD", "Call Record"
        CALENDAR = "CALENDAR", "Calendar"

    is_enable = models.BooleanField(default=True)
    subscription_id = models.CharField(max_length=255, null=True, blank=True)
    change_type = models.CharField(max_length=255, null=True, blank=True)
    notification_url = models.CharField(max_length=255, null=True, blank=True)
    resource = models.CharField(max_length=255, null=True, blank=True)
    type_resource = models.CharField(
        max_length=32,
        choices=TypeResource.choices,
        default=TypeResource.CALLRECORD,
        verbose_name="Type of the resource",
    )
    tenant_user = models.ForeignKey(
        "wbintegrator_office365.TenantUser",
        related_name="subscriptions",
        null=True,
        blank=True,
        on_delete=models.deletion.SET_NULL,
        verbose_name="Tenant User",
    )
    expiration_date = models.DateTimeField(null=True, blank=True, verbose_name="Expiration Date")
    created = models.DateTimeField(auto_now_add=True, verbose_name="Created")
    application_id = models.CharField(null=True, blank=True, max_length=255)
    creator_id = models.CharField(null=True, blank=True, max_length=255)
    client_state = models.CharField(null=True, blank=True, max_length=255)
    latest_supported_tls_version = models.CharField(null=True, blank=True, max_length=255)
    notification_content_type = models.CharField(null=True, blank=True, max_length=255)
    odata_context = models.CharField(null=True, blank=True, max_length=255)
    encryption_certificate_id = models.CharField(null=True, blank=True, max_length=255)
    encryption_certificate = models.CharField(null=True, blank=True, max_length=255)
    include_resource_data = models.CharField(null=True, blank=True, max_length=255)
    notification_query_options = models.CharField(null=True, blank=True, max_length=255)

    def __str__(self):
        return f"{self.subscription_id}: {self.expiration_date}"

    def resubscribe(self) -> None:
        if self.subscription_id and (data := MicrosoftGraphAPI()._renew_subscription(self.subscription_id)):
            Subscription.objects.filter(id=self.id).update(
                expiration_date=data.get("expiration"), notification_url=data.get("notification_url")
            )

    @classmethod
    def get_endpoint_basename(self):
        return "wbintegrator_office365:subscription"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbintegrator_office365:subscriptionrepresentation-list"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{subscription_id}}: {{expiration_date}}"


@shared_task
def subscribe(instance_id, resource, minutes=4230, change_type="created, updated"):
    with transaction.atomic():
        data = MicrosoftGraphAPI()._subscribe(resource, minutes, change_type=change_type)
        if data.get("id"):
            instance = Subscription.objects.get(id=instance_id)
            if instance.type_resource == Subscription.TypeResource.CALLRECORD:
                instance.tenant_user = None
            instance.subscription_id = data.get("id")
            instance.change_type = data.get("change_type")
            instance.notification_url = data.get("notification_url")
            instance.resource = data.get("resource")
            instance.expiration_date = data.get("expiration")
            instance.application_id = data.get("application_id")
            instance.creator_id = data.get("creator_id")
            instance.client_state = data.get("client_state")
            instance.latest_supported_tls_version = data.get("latest_supported_tls_version")
            instance.notification_content_type = data.get("notification_content_type")
            instance.odata_context = data.get("odata_context")
            instance.encryption_certificate_id = data.get("encryption_certificate_id")
            instance.encryption_certificate = data.get("encryption_certificate")
            instance.include_resource_data = data.get("include_resource_data")
            instance.notification_query_options = data.get("notification_query_options")
            instance.save()


@shared_task
def unsubscribe(subscription_id):
    if subscription_id:
        MicrosoftGraphAPI()._unsubscribe(subscription_id)


@shared_task
def verification_subscriptions(instance_id: int):
    with transaction.atomic():
        ms_subscriptions = [elt.get("id") for elt in MicrosoftGraphAPI().subscriptions()]
        qs_subscriptions = Subscription.objects.filter(Q(id=instance_id) & Q(subscription_id__isnull=False))
        qs_subscriptions.filter(~Q(subscription_id__in=ms_subscriptions)).update(is_enable=False)


@shared_task
def resubscribe_as_task(instance_id: int):
    subscription = Subscription.objects.get(id=instance_id)
    subscription.resubscribe()


@receiver(post_save, sender=Subscription)
def post_save_subscription(sender, instance, created, **kwargs):
    if (created and instance.is_enable) or (instance.is_enable and not instance.subscription_id):
        # Subscription to Call Event
        if instance.type_resource == Subscription.TypeResource.CALLRECORD and not instance.expiration_date:
            transaction.on_commit(
                lambda: chain(
                    subscribe.si(instance.id, "/communications/callRecords"),
                    verification_subscriptions.si(instance.id),
                ).apply_async()
            )
    elif not instance.is_enable and instance.subscription_id:
        transaction.on_commit(
            lambda: chain(
                unsubscribe.si(instance.subscription_id), verification_subscriptions.si(instance.id)
            ).apply_async()
        )


@receiver(post_delete, sender=Subscription)
def post_delete_subscription(sender, instance, **kwargs):
    unsubscribe.delay(instance.subscription_id)



from django.urls import include, path
from wbcore.routers import WBCoreRouter

from wbintegrator_office365.viewsets import viewsets

router = WBCoreRouter()
router.register(
    r"tenantuserrepresentation", viewsets.TenantUserRepresentationViewSet, basename="tenantuserrepresentation"
)
router.register(r"tenantuser", viewsets.TenantUserViewSet)

router.register(
    r"subscriptionrepresentation", viewsets.SubscriptionRepresentationViewSet, basename="subscriptionrepresentation"
)
router.register(r"subscription", viewsets.SubscriptionViewSet, basename="subscription")

router.register(r"eventrepresentation", viewsets.EventRepresentationViewSet, basename="eventrepresentation")
router.register(r"event", viewsets.EventViewSet, basename="event")

router.register(r"calluserrepresentation", viewsets.CallUserRepresentationViewSet, basename="calluserrepresentation")
router.register(r"calluser", viewsets.CallUserViewSet, basename="calluser")

router.register(
    r"calleventrepresentation", viewsets.CallEventRepresentationViewSet, basename="calleventrepresentation"
)
router.register(r"callevent", viewsets.CallEventViewSet, basename="callevent")

router.register(r"calleventchart", viewsets.CallEventReceptionTime, basename="calleventchart")
router.register(r"calleventsummarygraph", viewsets.CallEventSummaryGraph, basename="calleventsummarygraph")

router.register(r"eventlog", viewsets.EventLogViewSet, basename="eventlog")
router.register(r"eventlog", viewsets.EventLogRepresentationViewSet, basename="eventlogrepresentation")

event_router = WBCoreRouter()
event_router.register(
    r"eventlog",
    viewsets.EventLogEventViewSet,
    basename="event-eventlog",
)


urlpatterns = [
    path("", include(router.urls)),
    path("event/<str:last_event_id>/", include(event_router.urls)),
    path("listen", viewsets.listen, name="listen"),
    path("callback/permissions", viewsets.callback_consent_permission),
]



import pandas as pd


def parse(_dict, scalar_value=False):
    data = None
    if scalar_value:
        df = pd.DataFrame(_dict, index=[0])
    else:
        df = pd.DataFrame(_dict)
    df = df.rename(columns=_map)
    data = df.to_dict("records")
    return data


_map = {
    "lastModifiedDateTime": "last_modified",
    "startDateTime": "start",
    "endDateTime": "end",
    "organizer": "organizer",
    "displayName": "display_name",
    "businessPhones": "business_phones",
    "mobilePhone": "mobile_phone",
    "surname": "surname",
    "givenName": "given_name",
    "mailNickname": "mail_nickname",
    "userPrincipalName": "user_principal_name",
    "notificationUrl": "notification_url",
    "changeType": "change_type",
    "expirationDateTime": "expiration",
    "resourceData": "resource_data",
    "subscriptionExpirationDateTime": "subscription_expiration",
    "clientState": "client_state",
    "tenantId": "tenant_id",
    "subscriptionId": "subscription_id",
    "organizer.user": "organizer.user",
    "organizer.user.id": "organizer.user.id",
    "organizer.user.displayName": "organizer.user.display_name",
    "organizer.phone": "organizer.phone",
    "organizer.phone.id": "organizer.phone.id",
    "organizer.phone.displayName": "organizer.phone.display_name",
    "organizer.guest": "organizer.guest",
    "organizer.guest.id": "organizer.guest.id",
    "organizer.guest.displayName": "organizer.guest.display_name",
    "organizer.spoolUser": "organizer.splool_user",
    "organizer.acsUser": "organizer.acs_user",
    "organizer.encrypted": "organizer.encrypted",
    "organizer.onPremises": "organizer.on_premises",
    "organizer.acsApplicationInstance": "organizer.acs_appllication_instance",
    "organizer.spoolApplicationInstance": "organizer.spool_application_instance",
    "organizer.applicationInstance": "organizer.application_instance",
    "organizer.application": "organizer.application",
    "organizer.device": "organizer.device",
    "joinWebUrl": "join_web_url",
    "preferredLanguage": "preferred_language",
    "officeLocation": "office_location",
    "jobTitle": "job_title",
    "createdDateTime": "created",
    "isReminderOn": "is_reminder_on",
    "reminderMinutesBeforeStart": "reminder_minutes_before_start",
    "hasAttachments": "has_attachments",
    "webLink": "web_link",
    "start.dateTime": "start",
    "start.timeZone": "start.time_zone",
    "end.dateTime": "end",
    "end.timeZone": "end.time_zone",
    "location.displayName": "location.display_name",
    "location.locationUri": "location.location_uri",
    "location.locationType": "location.location_type",
    "location.uniqueId": "location.unique_id",
    "location.uniqueIdType": "location.unique_id_type",
    "location.address.type": "location.address.type",
    "location.address.street": "location.address.street",
    "location.address.city": "location.address.city",
    "location.address.postalCode": "location.address.postal_code",
    "location.address.countryOrRegion": "location.address.country_or_region",
    "recurrence.pattern.daysOfWeek": "recurrence.pattern.days_of_week",
    "recurrence.pattern.dayOfMonth": "recurrence.pattern.day_of_month",
    "recurrence.pattern.firstDayOfWeek": "recurrence.pattern.first_day_of_week",
    "recurrence.range.recurrenceTimeZone": "recurrence.range.recurrence_time_zone",
    "recurrence.range.numberOfOccurrences": "recurrence.range.number_of_occurrences",
    "recurrence.range.startDate": "recurrence.range.start_date",
    "recurrence.range.endDate": "recurrence.range.end_date",
    "organizer.emailAddress.name": "organizer.email_address.name",
    "organizer.emailAddress.address": "organizer.email_address.address",
    "responseStatus.response": "response_status.response",
    "responseStatus.time": "response_status.time",
    "attendees": "participants",
    "@odata.context": "odata_context",
    "@odata.etag": "odata_etag",
    "applicationId": "application_id",
    "includeResourceData": "include_resource_data",
    "latestSupportedTlsVersion": "latest_supported_tls_version",
    "encryptionCertificate": "encryption_certificate",
    "encryptionCertificateId": "encryption_certificate_id",
    "notificationQueryOptions": "notification_query_options",
    "notificationContentType": "notification_content_type",
    "lifecycleNotificationUrl": "lifecycle_notification_url",
    "creatorId": "creator_id",
    "isDefaultCalendar": "is_default_calendar",
    "hexColor": "hex_color",
    "changeKey": "change_key",
    "canShare": "can_share",
    "canViewPrivateItems": "can_view_private_items",
    "isShared": "is_shared",
    "isSharedWithMe": "is_shared_with_me",
    "canEdit": "can_edit",
    "allowedOnlineMeetingProviders": "allowed_online_meeting_providers",
    "defaultOnlineMeetingProvider": "default_online_meeting_provider",
    "isTallyingResponses": "is_tallying_responses",
    "isRemovable": "is_removable",
    "teamsForBusiness": "teams_for_business",
    "transactionId": "transaction_id",
    "originalStartTimeZone": "original_start_time_zone",
    "originalEndTimeZone": "original_end_time_zone",
    "bodyPreview": "body_preview",
    "isAllDay": "is_all_day",
    "isCancelled": "is_cancelled",
    "isOrganizer": "is_organizer",
    "responseRequested": "response_requested",
    "seriesMasterId": "series_master_id",
    "showAs": "show_as",
    "onlineMeeting": "online_meeting",
    "onlineMeetingUrl": "online_meeting_url",
    "isOnlineMeeting": "is_online_meeting",
    "onlineMeetingProvider": "online_meeting_provider",
    "allowNewTimeProposals": "allow_new_time_proposals",
    "occurrenceId": "occurrence_id",
    "isDraft": "is_draft",
    "hideAttendees": "hide_attendees",
    "responseStatus": "response_status",
    "appId": "app_id",
    "passwordCredentials": "password_credentials",
    "iCalUId": "uid",
    "body.contentType": "body.content_type",
}



import math
from datetime import date, timedelta

from celery import shared_task
from django.contrib.auth import get_user_model
from django.db.models import Q
from wbcore.contrib.directory.models import Person
from wbcore.contrib.notifications.dispatch import send_notification

from wbintegrator_office365.importer import MicrosoftGraphAPI
from wbintegrator_office365.models.event import CallEvent
from wbintegrator_office365.models.subscription import Subscription


@shared_task
def send_call_summary(to_emails: list, profile_ids: list, offset: int = 1):
    for to_email in to_emails:
        if (recipient := get_user_model().objects.filter(email=to_email).first()) and (
            profiles := Person.objects.filter(id__in=profile_ids)
        ):
            _day = date.today() - timedelta(days=offset)

            message = "<html><head><style> #summery_table tr:nth-child(even){background-color: #f2f2f2;}</style></head><body>"

            for profile in profiles:
                call_events = CallEvent.objects.filter(
                    participants__tenant_user__profile__computed_str__icontains=profile.computed_str,
                    start__date=_day,
                    end__date=_day,
                ).order_by("start")

                message += f"<p> \
                                <span><b> {profile.computed_str} </b></span><br/> \
                                <span>Date: <b>{_day}</b> </span><br/> \
                                <span>Total number of calls: <b>{call_events.count()}</b> </span> \
                            </p>"
                if call_events.count():
                    message += "<table id='summery_table' style='border-collapse: collapse;'> \
                                <tr style='color: white; background-color: #1868ae;'> \
                                    <th style='border: 1px solid #ddd;padding: 10px 7px;' >Start</th> \
                                    <th style='border: 1px solid #ddd;padding: 10px 7px;' >End</th> \
                                    <th style='border: 1px solid #ddd;padding: 10px 7px;' >Duration (min)</th> \
                                    <th style='border: 1px solid #ddd;padding: 10px 7px;' >Organized by</th> \
                                    <th style='border: 1px solid #ddd;padding: 10px 7px;' >Participants</th> \
                                </tr>"
                    for call in call_events:
                        participants = ""
                        count = 0
                        total_participants = call.participants.exclude(tenant_user__profile=profile).count()
                        for participant in call.participants.exclude(tenant_user__profile=profile):
                            count += 1
                            participants += f"{participant.__str__()}"
                            if count < total_participants:
                                participants += ", "

                        delta_in_minutes = divmod((call.end - call.start).total_seconds(), 60)
                        _duration = f"{math.floor(delta_in_minutes[0])}:{math.floor(delta_in_minutes[1])}"
                        message += f"<tr> \
                                    <td style='border: 1px solid #ddd;padding: 2px;' >{call.start.astimezone().strftime('%Y-%m-%d %H:%M:%S')}</td> \
                                    <td style='border: 1px solid #ddd;padding: 2px;' >{call.end.astimezone().strftime('%Y-%m-%d %H:%M:%S')}</td> \
                                    <td style='border: 1px solid #ddd;padding: 0px;' text-align:center;><b>{_duration}</b></td> \
                                    <td style='border: 1px solid #ddd;padding: 2px;' ><b>{call.organizer.__str__()}</b></td> \
                                    <td style='border: 1px solid #ddd;padding: 2px;' >{participants}</td> \
                                </tr>"
                    message += "</table><br/>"

            message += "</body></html>"

        send_notification(
            code="wbintegrator_office365.callevent.notify",
            title=f"Call summary - {_day}",
            body=message,
            user=recipient,
        )


@shared_task
def notify_no_active_call_record_subscription(to_email):
    recipient = get_user_model().objects.filter(email=to_email)
    ms_subscriptions = [elt.get("id") for elt in MicrosoftGraphAPI().subscriptions()]
    qs_subscriptions = Subscription.objects.filter(
        Q(is_enable=True) & Q(subscription_id__isnull=False) & Q(type_resource=Subscription.TypeResource.CALLRECORD)
    )
    enable_subcriptions = qs_subscriptions.filter(subscription_id__in=ms_subscriptions)
    if recipient.exists() and (
        len(ms_subscriptions) == 0 or (qs_subscriptions.count() > 0 and enable_subcriptions.count() == 0)
    ):
        _day = date.today()
        send_notification(
            code="wbintegrator_office365.callevent.notify",
            title=f"No active Call Record subscriptions in Microsoft - {_day}",
            body=f"""<p>There are currently no active Call record subscriptions in Microsoft, so we are no longer receiving calls, Please check</p>
            <ul>
                <li>Number of subscriptions on Microsoft: <b>{len(ms_subscriptions)}</b></li>
                <li>Number of Call subscriptions: <b>{qs_subscriptions.count()}</b></li>
                <li>Number of enabled calling subscriptions: <b>{enable_subcriptions.count()}</b></li>

            </ul>
            """,
            user=recipient.first(),
        )


@shared_task
def periodic_resubscribe_task():
    for subscription in Subscription.objects.filter(
        is_enable=True, type_resource=Subscription.TypeResource.CALLRECORD, subscription_id__isnull=False
    ):
        subscription.resubscribe()
